import type { INestApplication } from '@nestjs/common';
import type { ClientProxy } from '@nestjs/microservices';
import type {
  EventAnonCredsCredentialRequestGetAllInput,
  EventAnonCredsCredentialsGetAllInput,
  EventAnonCredsCredentialsGetByIdInput,
  EventDidcommAnonCredsCredentialsOfferToSelfInput,
  EventAnonCredsCredentialOfferGetAllInput,
  EventAnonCredsCredentialOfferGetByIdInput,
  EventAnonCredsCredentialRequestGetByIdInput,
  EventAnonCredsCredentialsDeleteByIdInput,
} from '@ocm/shared';

import { AutoAcceptCredential, CredentialExchangeRecord } from '@credo-ts/core';
import { ClientsModule, Transport } from '@nestjs/microservices';
import { Test } from '@nestjs/testing';
import {
  EventAnonCredsCredentialsDeleteById,
  EventAnonCredsCredentialOfferGetAll,
  EventAnonCredsCredentialOfferGetById,
  EventAnonCredsCredentialRequestGetAll,
  EventAnonCredsCredentialRequestGetById,
  EventAnonCredsCredentialsGetAll,
  EventAnonCredsCredentialsGetById,
  EventAnonCredsProofsDeleteById,
  EventDidcommAnonCredsCredentialsOfferToSelf,
} from '@ocm/shared';
import { randomBytes } from 'crypto';
import { firstValueFrom } from 'rxjs';

import { AgentModule } from '../src/agent/agent.module.js';
import { AnonCredsCredentialsModule } from '../src/agent/anoncredsCredentials/anoncredsCredentials.module.js';
import { ConnectionsModule } from '../src/agent/connections/connections.module.js';
import { ConnectionsService } from '../src/agent/connections/connections.service.js';
import { CredentialDefinitionsModule } from '../src/agent/credentialDefinitions/credentialDefinitions.module.js';
import { CredentialDefinitionsService } from '../src/agent/credentialDefinitions/credentialDefinitions.service.js';
import { DidsModule } from '../src/agent/dids/dids.module.js';
import { DidsService } from '../src/agent/dids/dids.service.js';
import { SchemasModule } from '../src/agent/schemas/schemas.module.js';
import { SchemasService } from '../src/agent/schemas/schemas.service.js';
import { TenantsModule } from '../src/agent/tenants/tenants.module.js';
import { TenantsService } from '../src/agent/tenants/tenants.service.js';
import { mockConfigModule } from '../src/config/__tests__/mockConfig.js';

describe('Credentials', () => {
  const TOKEN = 'CREDENTIALS_CLIENT_SERVICE';
  let app: INestApplication;
  let client: ClientProxy;
  let tenantId: string;

  let issuerDid: string;
  let credentialDefinitionId: string;

  beforeAll(async () => {
    const moduleRef = await Test.createTestingModule({
      imports: [
        mockConfigModule(3004, true),
        AgentModule,
        ConnectionsModule,
        SchemasModule,
        CredentialDefinitionsModule,
        AnonCredsCredentialsModule,
        TenantsModule,
        DidsModule,
        ClientsModule.register([{ name: TOKEN, transport: Transport.NATS }]),
      ],
    }).compile();

    app = moduleRef.createNestApplication();

    app.connectMicroservice({ transport: Transport.NATS });

    await app.startAllMicroservices();
    await app.init();

    client = app.get(TOKEN);
    await client.connect();

    const tenantsService = app.get(TenantsService);
    const { id } = await tenantsService.create({ label: TOKEN });
    tenantId = id;

    const connectionsService = app.get(ConnectionsService);
    await connectionsService.createConnectionWithSelf({ tenantId });

    const ds = app.get(DidsService);
    await ds.registerEndorserDids();
    const [did] = await ds.registerDidIndyFromSeed({
      tenantId,
      seed: randomBytes(16).toString('hex'),
    });
    issuerDid = did;

    const schemaService = app.get(SchemasService);
    const { schemaId } = await schemaService.register({
      issuerDid,
      tenantId,
      name: 'test-schema-name',
      version: `1.${Date.now()}`,
      attributeNames: ['Name', 'Age'],
    });

    const credentialDefinitionService = app.get(CredentialDefinitionsService);
    const { credentialDefinitionId: cdi } =
      await credentialDefinitionService.register({
        supportsRevocation: false,
        tenantId,
        issuerDid,
        schemaId,
        tag: `default-${Date.now()}`,
      });

    credentialDefinitionId = cdi;
  });

  afterAll(async () => {
    await app.close();
    client.close();
  });

  it(EventAnonCredsCredentialsGetAll.token, async () => {
    const response$ = client.send<
      EventAnonCredsCredentialsGetAll,
      EventAnonCredsCredentialsGetAllInput
    >(EventAnonCredsCredentialsGetAll.token, { tenantId });
    const response = await firstValueFrom(response$);
    const eventInstance = EventAnonCredsCredentialsGetAll.fromEvent(response);

    expect(eventInstance.instance).toEqual(expect.arrayContaining([]));
  });

  it(EventAnonCredsCredentialOfferGetAll.token, async () => {
    const response$ = client.send<
      EventAnonCredsCredentialOfferGetAll,
      EventAnonCredsCredentialOfferGetAllInput
    >(EventAnonCredsCredentialOfferGetAll.token, { tenantId });
    const response = await firstValueFrom(response$);
    const eventInstance =
      EventAnonCredsCredentialOfferGetAll.fromEvent(response);

    expect(eventInstance.instance).toEqual(expect.arrayContaining([]));
  });

  it(EventAnonCredsCredentialOfferGetById.token, async () => {
    const response$ = client.send<
      EventAnonCredsCredentialOfferGetById,
      EventAnonCredsCredentialOfferGetByIdInput
    >(EventAnonCredsCredentialOfferGetById.token, {
      tenantId,
      credentialOfferId: 'some-id-that-does-not-exist',
    });
    const response = await firstValueFrom(response$);
    const eventInstance =
      EventAnonCredsCredentialOfferGetById.fromEvent(response);

    expect(eventInstance.instance).toBeNull();
  });

  it(EventAnonCredsCredentialRequestGetAll.token, async () => {
    const response$ = client.send<
      EventAnonCredsCredentialRequestGetAll,
      EventAnonCredsCredentialRequestGetAllInput
    >(EventAnonCredsCredentialRequestGetAll.token, { tenantId });
    const response = await firstValueFrom(response$);
    const eventInstance =
      EventAnonCredsCredentialRequestGetAll.fromEvent(response);

    expect(eventInstance.instance).toEqual(expect.arrayContaining([]));
  });

  it(EventAnonCredsCredentialRequestGetById.token, async () => {
    const response$ = client.send<
      EventAnonCredsCredentialRequestGetById,
      EventAnonCredsCredentialRequestGetByIdInput
    >(EventAnonCredsCredentialRequestGetById.token, {
      tenantId,
      credentialRequestId: 'some-id-that-does-not-exist',
    });
    const response = await firstValueFrom(response$);
    const eventInstance =
      EventAnonCredsCredentialRequestGetById.fromEvent(response);

    expect(eventInstance.instance).toBeNull();
  });

  it(EventAnonCredsCredentialsGetById.token, async () => {
    const response$ = client.send<
      EventAnonCredsCredentialsGetById,
      EventAnonCredsCredentialsGetByIdInput
    >(EventAnonCredsCredentialsGetById.token, {
      tenantId,
      credentialRecordId: 'some-id',
    });
    const response = await firstValueFrom(response$);
    const eventInstance = EventAnonCredsCredentialsGetById.fromEvent(response);

    expect(eventInstance.instance).toEqual(null);
  });

  it(EventDidcommAnonCredsCredentialsOfferToSelf.token, async () => {
    const attributes = [
      { name: 'Name', value: 'Berend' },
      { name: 'Age', value: '25' },
    ];

    const response$ = client.send<
      EventDidcommAnonCredsCredentialsOfferToSelf,
      EventDidcommAnonCredsCredentialsOfferToSelfInput
    >(EventDidcommAnonCredsCredentialsOfferToSelf.token, {
      tenantId,
      credentialDefinitionId,
      attributes,
    });

    const response = await firstValueFrom(response$);
    const eventInstance =
      EventDidcommAnonCredsCredentialsOfferToSelf.fromEvent(response);

    expect(eventInstance.instance).toMatchObject({
      autoAcceptCredential: AutoAcceptCredential.Always,
    });
  });

  it(EventAnonCredsProofsDeleteById.token, async () => {
    let credentialExchangeRecord: CredentialExchangeRecord | undefined =
      undefined;

    // GET ALL
    {
      const response$ = client.send<
        EventAnonCredsCredentialsGetAll,
        EventAnonCredsCredentialsGetAllInput
      >(EventAnonCredsCredentialsGetAll.token, { tenantId });
      const response = await firstValueFrom(response$);
      const eventInstance = EventAnonCredsCredentialsGetAll.fromEvent(response);

      expect(eventInstance.instance).toEqual(expect.arrayContaining([]));
    }

    // Offer a credential

    {
      const attributes = [
        { name: 'Name', value: 'Berend' },
        { name: 'Age', value: '25' },
      ];

      const response$ = client.send<
        EventDidcommAnonCredsCredentialsOfferToSelf,
        EventDidcommAnonCredsCredentialsOfferToSelfInput
      >(EventDidcommAnonCredsCredentialsOfferToSelf.token, {
        tenantId,
        credentialDefinitionId,
        attributes,
      });

      const response = await firstValueFrom(response$);
      const eventInstance =
        EventDidcommAnonCredsCredentialsOfferToSelf.fromEvent(response);

      expect(eventInstance.instance).toMatchObject({
        autoAcceptCredential: AutoAcceptCredential.Always,
      });

      credentialExchangeRecord = eventInstance.instance;
    }

    // GET THE CREDENTIAL BY ID
    {
      const response$ = client.send<
        EventAnonCredsCredentialsGetById,
        EventAnonCredsCredentialsGetByIdInput
      >(EventAnonCredsCredentialsGetById.token, {
        tenantId,
        credentialRecordId: credentialExchangeRecord.id,
      });
      const response = await firstValueFrom(response$);
      const eventInstance =
        EventAnonCredsCredentialsGetById.fromEvent(response);

      expect(eventInstance.instance).toBeInstanceOf(CredentialExchangeRecord);
    }

    // DELETE THE CREDENTIAL
    {
      const response$ = client.send<
        EventAnonCredsCredentialsDeleteById,
        EventAnonCredsCredentialsDeleteByIdInput
      >(EventAnonCredsCredentialsDeleteById.token, {
        tenantId,
        credentialRecordId: credentialExchangeRecord.id,
      });
      const response = await firstValueFrom(response$);
      const eventInstance =
        EventAnonCredsCredentialsDeleteById.fromEvent(response);

      expect(eventInstance).toMatchObject({});
    }

    // GET THE CREDENTIAL BY ID
    {
      const response$ = client.send<
        EventAnonCredsCredentialsGetById,
        EventAnonCredsCredentialsGetByIdInput
      >(EventAnonCredsCredentialsGetById.token, {
        tenantId,
        credentialRecordId: credentialExchangeRecord.id,
      });
      const response = await firstValueFrom(response$);
      const eventInstance =
        EventAnonCredsCredentialsGetById.fromEvent(response);

      expect(eventInstance.instance).toBeNull();
    }
  });
});
